/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */
/**
 *  \brief The core system of the benchmarking suite: compute configurations
 *         and perform the benchmark
 *         
 *
 *  \author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */
 
#ifndef POSFINDER_H
#define POSFINDER_H

/*
 * Includes ;
 */
#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNodePtr.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGGeometry.h>
#include <vector>
#include "Settings.h"
#include "Utils.h"

/**
 *  \class PosFinder ;
 *  \brief implements the benchmarking core system
 */
class   PosFinder
{

private:
    double          x_deg, y_deg, z_deg;
    osg::NodePtr    scene;  /* OpenSG root node */
    osg::Matrix     matrix; /* configuration matrix */

    void            saveQueue( const char *name, double dist ); /* Save
                                                                 * configurations
                                                                 * in queue */
    void            loadQueue( const char *name );  /* load configurations from
                                                     * queue */

protected:
    Settings            *settings;  /* parameters for benchmarking*/
    osg::TransformPtr   movingTrf; /*nodes transformation node */
    osg::NodePtr        fixedNode, movingNode, trfNode, distNode;
    osg::Vec3f          fixedCenter, movingCenter; /* Center of objects */
    osg::Vec3f          fixedSize, movingSize; /* size of objects */
    NanoTimer           nanotimer; /* used for exact timing */
    double              buildTime; /* Store the time needed for building the BVHs */

    void                setSize();
    bool                testBetween( const osg::Pnt3f & pos1,
                                     const osg::Pnt3f & pos2,
                                     osg::Pnt3f & ret );
    void                queueBetween( const osg::Pnt3f    &pos1,
                                      const osg::Pnt3f    &pos2 );
    double              testPos( const osg::Pnt3f &pos );
    bool                testBetweenZero( const osg::Pnt3f & pos1,
                                         const osg::Pnt3f & pos2,
                                         osg::Pnt3f & ret );
    bool                testBin( const osg::Pnt3f &pos );
    void                queue( const osg::Pnt3f &pos );
    void                workQueue();

public:
    std::vector<osg::Matrix>    testQueue; /* queue with configurations to test*/
    PosFinder( Settings *set ); /* initialize PosFinder with parameters from settings */
    osg::NodePtr                getScene(); /* returns root node */
    void                        testAll(); /* start benchmark of all configurations */
    bool                        doRot(); /* performs a single rotation of the moving object */
    virtual bool                iterate() = 0; /* performs benchmark for a single configuration */
    virtual bool                rotate() = 0; /* rotate mocing object by values specified in *_deg */
    void                        printBottom(); /* print some information about benchmark */
};
#endif
