/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ Opcode \
 */

/**
 *  @class Wrapper for Opcode
 *  @author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
 
#include <vector>
#include <stdio.h>

#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNode.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGBoxVolume.h>

#include <Stdafx.h>

#include "../Bench.h"

/*
 * Defines ;
 */
 
#define getGeom( x )    ( osg::GeometryPtr::dcast( (x)->getCore() ) )

using namespace                     std;
using osg::                         Vec3f;
using osg::                         Pnt3f;

/*
 * Forward Declarations ;
 */
 
static void                         makeObject( int num, osg::NodePtr Node[] );
static int                          doCollide( int num, osg::Matrix m[] );
static void                         shutdown();

static tAPI                         API = { makeObject, doCollide, NULL, shutdown };

vector<IceMaths::Point>             *Points;
vector<IceMaths::IndexedTriangle>   *Tri;
static Opcode::MeshInterface        **MI;
Opcode::Model **Obj;
Opcode::AABBTreeCollider    TC;
Opcode::BVTCache            ColCache;

/* -----------------------------------------------------------------------------
 *  \brief Init module and register it to the benchmarking application
 ------------------------------------------------------------------------------- */
 
#ifdef _WIN32
extern "C"
__declspec( dllexport )
std::vector<Algorithm *> *
initModule ( void )
#else
extern "C"
{
std::vector<Algorithm *>    *initModule( void );
}
std::vector<Algorithm *>    *
initModule ( void )
#endif
{
    std::vector<Algorithm *>    *ret = new std::vector < Algorithm * > ();
    ret->clear();
    Algorithm   *opcode = new Algorithm();
    opcode->shortName = "op";
    opcode->name = "Opcode";
    opcode->api = API;
    ret->push_back( opcode );

    TC.SetFirstContact( true );

    return ret;
}


/* -----------------------------------------------------------------------------
 *  \brief Collision check
 *  \param m transformation matrix of moving object
 *  \param i number of transformed object (has no effect in this version)
 ------------------------------------------------------------------------------- */
static int doCollide( int num, osg::Matrix m[] )
{
    IceMaths::Matrix4x4 transMatrix( m[0][0][0], m[0][0][1], m[0][0][2],
                                     m[0][0][3], m[0][1][0], m[0][1][1],
                                     m[0][1][2], m[0][1][3], m[0][2][0],
                                     m[0][2][1], m[0][2][2], m[0][2][3],
                                     m[0][3][0], m[0][3][1], m[0][3][2],
                                     m[0][3][3] );

    TC.Collide( ColCache, &transMatrix );
    return TC.GetContactStatus();
}

static Opcode::MeshInterface *
addObject( osg::GeometryPtr geo, int num )
{
    osg::TriangleIterator   ti = geo->
    beginTriangles(),
    ti_end = geo->endTriangles();
    osg::MFPnt3f * P = osg::GeoPositions3fPtr::dcast( geo->getPositions() )->getFieldPtr();
    Opcode::MeshInterface * MI = new Opcode::MeshInterface;
    unsigned    Tris = 0;

    for( unsigned i = 0; i < P->size(); ++i )
    {
        Points[num].push_back( IceMaths::Point(double((*P)[i][0]), (*P)[i][1], (*P)[i][2]) );
    }
    for( ; ti != ti_end; ++ti )
    {
        Tri[num].push_back
            (
                IceMaths::IndexedTriangle
                    (
                        ti.getPositionIndex(0), ti.getPositionIndex
                            (1), ti.getPositionIndex(2)
                    )
            );
        ++Tris;
    }
    MI->SetNbVertices( P->size() );
    MI->SetNbTriangles( Tris );
    MI->
    SetPointers( &Tri[num][0], &Points[num][0] ),
    Tris, P->size();
    return MI;
}

/* -----------------------------------------------------------------------------
 *  \brief Make the OSG-nodes collidable
 *  \param num # of objects
 *  \param Node OSG-Nodes of the objects
 ------------------------------------------------------------------------------- */
static void makeObject( int num, osg::NodePtr Node[] )
{
    Points = new vector < IceMaths::Point >[num];
    Tri = new vector < IceMaths::IndexedTriangle >[num];
    MI = new Opcode::MeshInterface *[num + 1];
    MI[num] = NULL;
    Obj = new Opcode::Model *[num];
    for( int i = 0; i < num; ++i )
    {
        Opcode::OPCODECREATE    OPCC;

        Obj[i] = new Opcode::Model;
        OPCC.mIMesh = MI[i] = addObject( getGeom( Node[i] ), i );
        Obj[i]->Build( OPCC );
    }

    ColCache.Model0 = Obj[0];
    ColCache.Model1 = Obj[1];

    //
    // TC.SetPointers0(&Tri[0][0], &Points[0][0]);
    // TC.SetPointers1(&Tri[1][0], &Points[1][0]);
    //
}

static void shutdown()
{
    for ( int i = 0; MI[i]; ++i )
    {
        delete Obj[i];
        delete MI[i];
    }
    delete[] MI;
    MI = NULL;
    delete[] Obj;
    Obj = NULL;
    delete[] Tri;
    Tri = NULL;
    delete[] Points;
    Points = NULL;
}
