/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ VCollide \
 */

/**
 *  @class Wrapper for VCollide
 *  @author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
 
#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNode.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGBoxVolume.h>

#include <VCollide.H>

#include "../Bench.h"

/*
 * Defines ;
 */
 
#define getGeom( x )    ( osg::GeometryPtr::dcast( (x)->getCore() ) )

using osg::Vec3f;
using osg::Pnt3f;

/*
 * Forward Declarations ;
 */
 
VCollide                    *VC;
int                         *vcObjs;

static void                 makeVCObject( int num, osg::NodePtr Node[] );
static int                  doVCollide( int num, osg::Matrix m[] );
static void                 shutdown();

static tAPI                 API = { makeVCObject, doVCollide, NULL, shutdown };


/* -----------------------------------------------------------------------------
 *  \brief Init module and register it to the benchmarking application
 ------------------------------------------------------------------------------- */
#ifdef _WIN32
extern "C"
__declspec( dllexport )
std::vector<Algorithm *> *
initModule ( void )
#else
extern "C"
{
std::vector<Algorithm *>    *initModule( void );
}
std::vector<Algorithm *>    *
initModule ( void )
#endif
{
    std::vector<Algorithm *>    *ret = new std::vector < Algorithm * > ();
    ret->clear();
    Algorithm   *vcol = new Algorithm();
    vcol->shortName = "vc";
    vcol->name = "VCollide";
    vcol->api = API;
    ret->push_back( vcol );

    return ret;
}

/* -----------------------------------------------------------------------------
 *  \brief Collision check
 *  \param m transformation matrix of moving object
 *  \param i number of transformed object (has no effect in this version)
 ------------------------------------------------------------------------------- */
static int doVCollide( int num, osg::Matrix m[] )
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    double  transMatrix[4][4] = { { m[0][0][0], m[0][1][0], m[0][2][0], m[0][3][0] }, { m[0][0][1], m[0][1][1], m[0][2][1], m[0][3][1] }, { m[0][0][2], m[0][1][2], m[0][2][2], m[0][3][2] }, { m[0][0][3], m[0][1][3], m[0][2][3], m[0][3][3] } };
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    VC->UpdateTrans( vcObjs[0], transMatrix );
    VC->Collide();
    return VC->Report( 0, 0 );
}

/* -----------------------------------------------------------------------------
 *  \brief Add an object to the collision detection library
 ------------------------------------------------------------------------------- */
static int addObject( VCollide *VC, osg::GeometryPtr geo )
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    int                     ret;
    osg::TriangleIterator   ti = geo->beginTriangles(),
    ti_end = geo->endTriangles();
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    VC->NewObject( &ret );
    for( ; ti != ti_end; ++ti )
    {
        double  PosArray[3][3];
        Pnt3f   Pos[3];
        for( int i = 0; i < 3; ++i )
        {
            Pos[i] = ti.getPosition( i );
            PosArray[i][0] = Pos[i][0];
            PosArray[i][1] = Pos[i][1];
            PosArray[i][2] = Pos[i][2];
        }

        VC->AddTri( PosArray[0], PosArray[1], PosArray[2] );
    }

    VC->EndObject();
    return ret;
}

/* -----------------------------------------------------------------------------
 *  \brief Make the OSG-nodes collidable
 *  \param num # of objects
 *  \param Node OSG-Nodes of the objects
 ------------------------------------------------------------------------------- */
static void makeVCObject( int num, osg::NodePtr Node[] )
{
    VC = new VCollide;
    vcObjs = new int[num];
    for( int i = 0; i < num; ++i )
    {
        vcObjs[i] = addObject( VC, getGeom( Node[i] ) );
    }
}

static void shutdown()
{
    if ( VC )
    {
        delete VC;
        VC = NULL;
    }
    if ( vcObjs )
    {
        delete[] vcObjs;
        vcObjs = NULL;
    }
}
